@@ -33,7 +33,6 @@
 #include <sys/stat.h>
 #include <sys/syscall.h>
 #include <sys/types.h>
-#include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
 
@@ -50,6 +49,8 @@
 
 #define MAX_BUF 1000
 
+static void sc_detach_views_of_writable(sc_distro distro, bool normal_mode);
+
 // TODO: simplify this, after all it is just a tmpfs
 // TODO: fold this into bootstrap
 static void setup_private_mount(const char *snap_name)
@@ -82,6 +83,9 @@ static void setup_private_mount(const char *snap_name)
 	sc_must_snprintf(base_dir, sizeof(base_dir), "/tmp/snap.%s", snap_name);
 	sc_must_snprintf(tmp_dir, sizeof(tmp_dir), "%s/tmp", base_dir);
 
+	/* Switch to root group so that mkdir and open calls below create filesystem
+	 * elements that are not owned by the user calling into snap-confine. */
+	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
 	// Create /tmp/snap.$SNAP_NAME/ 0700 root.root. Ignore EEXIST since we want
 	// to reuse and we will open with O_NOFOLLOW, below.
 	if (mkdir(base_dir, 0700) < 0 && errno != EEXIST) {
@@ -92,6 +96,13 @@ static void setup_private_mount(const char *snap_name)
 	if (base_dir_fd < 0) {
 		die("cannot open base directory %s", base_dir);
 	}
+	/* This seems redundant on first read but it has the non-obvious
+	 * property of changing existing directories  that have already existed
+	 * but had incorrect ownership or permission. This is possible due to
+	 * earlier bugs in snap-confine and due to the fact that some systems
+	 * use persistent /tmp directory and may not clean up leftover files
+	 * for arbitrarily long. This comment applies the following two pairs
+	 * of fchmod and fchown. */
 	if (fchmod(base_dir_fd, 0700) < 0) {
 		die("cannot chmod base directory %s to 0700", base_dir);
 	}
@@ -103,6 +114,7 @@ static void setup_private_mount(const char *snap_name)
 	if (mkdirat(base_dir_fd, "tmp", 01777) < 0 && errno != EEXIST) {
 		die("cannot create private tmp directory %s/tmp", base_dir);
 	}
+	(void)sc_set_effective_identity(old);
 	tmp_dir_fd = openat(base_dir_fd, "tmp",
 			    O_RDONLY | O_DIRECTORY | O_CLOEXEC | O_NOFOLLOW);
 	if (tmp_dir_fd < 0) {
@@ -148,7 +160,7 @@ static void setup_private_pts(void)
 	// copied from /etc/default/devpts
 	sc_do_mount("devpts", "/dev/pts", "devpts", MS_MGC_VAL,
 		    "newinstance,ptmxmode=0666,mode=0620,gid=5");
-	sc_do_mount("/dev/pts/ptmx", "/dev/ptmx", "none", MS_BIND, 0);
+	sc_do_mount("/dev/pts/ptmx", "/dev/ptmx", "none", MS_BIND, NULL);
 }
 
 struct sc_mount {
@@ -241,10 +253,10 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// replicate the state of the root filesystem into the scratch directory.
 	sc_do_mount(config->rootfs_dir, scratch_dir, NULL, MS_REC | MS_BIND,
 		    NULL);
-	// Make the scratch directory recursively private. Nothing done there will
-	// be shared with any peer group, This effectively detaches us from the
-	// original namespace and coupled with pivot_root below serves as the
-	// foundation of the mount sandbox.
+	// Make the scratch directory recursively slave. Nothing done there will be
+	// shared with the initial mount namespace. This effectively detaches us,
+	// in one way, from the original namespace and coupled with pivot_root
+	// below serves as the foundation of the mount sandbox.
 	sc_do_mount("none", scratch_dir, NULL, MS_REC | MS_SLAVE, NULL);
 	// Bind mount certain directories from the host filesystem to the scratch
 	// directory. By default mount events will propagate in both into and out
@@ -253,9 +265,14 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// disabling the "is_bidirectional" flag as can be seen below.
 	for (const struct sc_mount * mnt = config->mounts; mnt->path != NULL;
 	     mnt++) {
-		if (mnt->is_bidirectional && mkdir(mnt->path, 0755) < 0 &&
-		    errno != EEXIST) {
-			die("cannot create %s", mnt->path);
+
+		if (mnt->is_bidirectional) {
+			sc_identity old =
+			    sc_set_effective_identity(sc_root_group_identity());
+			if (mkdir(mnt->path, 0755) < 0 && errno != EEXIST) {
+				die("cannot create %s", mnt->path);
+			}
+			(void)sc_set_effective_identity(old);
 		}
 		sc_must_snprintf(dst, sizeof dst, "%s/%s", scratch_dir,
 				 mnt->path);
@@ -304,12 +321,33 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 		// Fixes the following bugs:
 		//  - https://bugs.launchpad.net/snap-confine/+bug/1580018
 		//  - https://bugzilla.opensuse.org/show_bug.cgi?id=1028568
-		const char *dirs_from_core[] =
-		    { "/etc/alternatives", "/etc/ssl", "/etc/nsswitch.conf",
+		const char *dirs_from_core[] = {
+			"/etc/alternatives", "/etc/nsswitch.conf",
+			// Some specific and privileged interfaces (e.g docker-support) give
+			// access to apparmor_parser from the base snap which at a minimum
+			// needs to use matching configuration from the base snap instead
+			// of from the users host system.
+			"/etc/apparmor", "/etc/apparmor.d",
+			// Use ssl certs from the base by default unless
+			// using Debian/Ubuntu classic (see below)
+			"/etc/ssl",
 			NULL
 		};
+
 		for (const char **dirs = dirs_from_core; *dirs != NULL; dirs++) {
 			const char *dir = *dirs;
+
+			// Special case for ubuntu/debian based
+			// classic distros that use the core* snap:
+			// here we use the host /etc/ssl
+			// to support custom ca-cert setups
+			if (sc_streq(dir, "/etc/ssl") &&
+			    config->distro == SC_DISTRO_CLASSIC &&
+			    sc_is_debian_like() &&
+			    sc_startswith(config->base_snap_name, "core")) {
+				continue;
+			}
+
 			if (access(dir, F_OK) != 0) {
 				continue;
 			}
@@ -384,9 +422,16 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 		// that we are re-execing from
 		char *src = NULL;
 		char self[PATH_MAX + 1] = { 0 };
-		if (readlink("/proc/self/exe", self, sizeof(self) - 1) < 0) {
+		ssize_t nread;
+		nread = readlink("/proc/self/exe", self, sizeof self - 1);
+		if (nread < 0) {
 			die("cannot read /proc/self/exe");
 		}
+		// Though we initialized self to NULs and passed one less to
+		// readlink, therefore guaranteeing that self is
+		// zero-terminated, perform an explicit assignment to make
+		// Coverity happy.
+		self[nread] = '\0';
 		// this cannot happen except when the kernel is buggy
 		if (strstr(self, "/snap-confine") == NULL) {
 			die("cannot use result from readlink: %s", self);
@@ -406,19 +451,23 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// the this directory on the host filesystem may not match the location in
 	// the desired root filesystem. In the "core" and "ubuntu-core" snaps the
 	// directory is always /snap. On the host it is a build-time configuration
-	// option stored in SNAP_MOUNT_DIR.
-	sc_must_snprintf(dst, sizeof dst, "%s/snap", scratch_dir);
-	sc_do_mount(SNAP_MOUNT_DIR, dst, NULL, MS_BIND | MS_REC, NULL);
-	sc_do_mount("none", dst, NULL, MS_REC | MS_SLAVE, NULL);
+	// option stored in SNAP_MOUNT_DIR. In legacy mode (or in other words, not
+	// in normal mode), we don't need to do this because /snap is fixed and
+	// already contains the correct view of the mounted snaps.
+	if (config->normal_mode) {
+		sc_must_snprintf(dst, sizeof dst, "%s/snap", scratch_dir);
+		sc_do_mount(SNAP_MOUNT_DIR, dst, NULL, MS_BIND | MS_REC, NULL);
+		sc_do_mount("none", dst, NULL, MS_REC | MS_SLAVE, NULL);
+	}
 	// Create the hostfs directory if one is missing. This directory is a part
 	// of packaging now so perhaps this code can be removed later.
-	if (access(SC_HOSTFS_DIR, F_OK) != 0) {
-		debug("creating missing hostfs directory");
-		if (mkdir(SC_HOSTFS_DIR, 0755) != 0) {
-			die("cannot perform operation: mkdir %s",
-			    SC_HOSTFS_DIR);
+	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
+	if (mkdir(SC_HOSTFS_DIR, 0755) < 0) {
+		if (errno != EEXIST) {
+			die("cannot perform operation: mkdir %s", SC_HOSTFS_DIR);
 		}
 	}
+	(void)sc_set_effective_identity(old);
 	// Ensure that hostfs isgroup owned by root. We may have (now or earlier)
 	// created the directory as the user who first ran a snap on a given
 	// system and the group identity of that user is visilbe on disk.
@@ -445,7 +494,7 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// code changes the nvidia code assumes it has access to the existing
 	// pre-pivot filesystem.
 	if (config->distro == SC_DISTRO_CLASSIC) {
-		sc_mount_nvidia_driver(scratch_dir);
+		sc_mount_nvidia_driver(scratch_dir, config->base_snap_name);
 	}
 	// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
 	//                    pivot_root
@@ -505,6 +554,43 @@ static void sc_bootstrap_mount_namespace(const struct sc_mount_config *config)
 	// mount table and software inspecting the mount table may become confused.
 	sc_must_snprintf(src, sizeof src, "%s/proc", SC_HOSTFS_DIR);
 	sc_do_umount(src, UMOUNT_NOFOLLOW | MNT_DETACH);
+	// Detach both views of /writable: the one from hostfs and the one directly
+	// visible in /writable. Interfaces don't grant access to this directory
+	// and it has a large duplicated view of many mount points.  Note that this
+	// is only applicable to ubuntu-core systems.
+	sc_detach_views_of_writable(config->distro, config->normal_mode);
+}
+
+static void sc_detach_views_of_writable(sc_distro distro, bool normal_mode)
+{
+	// Note that prior to detaching either mount point we switch the
+	// propagation to private to both limit the change to just this view and to
+	// prevent otherwise occurring event propagation from self-conflicting and
+	// returning EBUSY. A similar approach is used by snap-update-ns and is
+	// documented in umount(2).
+	const char *writable_dir = "/writable";
+	const char *hostfs_writable_dir = "/var/lib/snapd/hostfs/writable";
+
+	// Writable only exists on ubuntu-core.
+	if (distro == SC_DISTRO_CLASSIC) {
+		return;
+	}
+	// On all core distributions we see /var/lib/snapd/hostfs/writable that
+	// exposes writable, with a structure specific to ubuntu-core.
+	debug("detaching %s", hostfs_writable_dir);
+	sc_do_mount("none", hostfs_writable_dir, NULL,
+		    MS_REC | MS_PRIVATE, NULL);
+	sc_do_umount(hostfs_writable_dir, UMOUNT_NOFOLLOW | MNT_DETACH);
+
+	// On ubuntu-core 16, when the executed snap uses core as base we also see
+	// the /writable that we directly inherited from the initial mount
+	// namespace.
+	if (distro == SC_DISTRO_CORE16 && !normal_mode) {
+		debug("detaching %s", writable_dir);
+		sc_do_mount("none", writable_dir, NULL, MS_REC | MS_PRIVATE,
+			    NULL);
+		sc_do_umount(writable_dir, UMOUNT_NOFOLLOW | MNT_DETACH);
+	}
 }
 
 /**
@@ -564,7 +650,8 @@ static bool __attribute__((used))
 }
 
 void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
-			  const sc_invocation * inv)
+			  const sc_invocation * inv, const gid_t real_gid,
+			  const gid_t saved_gid)
 {
 	// Classify the current distribution, as claimed by /etc/os-release.
 	sc_distro distro = sc_classify_distro();
@@ -585,6 +672,7 @@ void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 			{"/var/tmp"},	// to get access to the other temporary directory
 			{"/run"},	// to get /run with sockets and what not
 			{"/lib/modules",.is_optional = true},	// access to the modules of the running kernel
+			{"/lib/firmware",.is_optional = true},	// access to the firmware of the running kernel
 			{"/usr/src"},	// FIXME: move to SecurityMounts in system-trace interface
 			{"/var/log"},	// FIXME: move to SecurityMounts in log-observe interface
 #ifdef MERGED_USR
@@ -626,10 +714,8 @@ void sc_populate_mount_ns(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 		sc_bootstrap_mount_namespace(&legacy_config);
 	}
 
-	// set up private mounts
 	// TODO: rename this and fold it into bootstrap
 	setup_private_mount(inv->snap_instance);
-
 	// set up private /dev/pts
 	// TODO: fold this into bootstrap
 	setup_private_pts();
@@ -672,7 +758,7 @@ void sc_ensure_shared_snap_mount(void)
 		// removed once we have a measurement and feedback mechanism that lets
 		// us decide based on measurable data.
 		sc_do_mount(SNAP_MOUNT_DIR, SNAP_MOUNT_DIR, "none",
-			    MS_BIND | MS_REC, 0);
+			    MS_BIND | MS_REC, NULL);
 		sc_do_mount("none", SNAP_MOUNT_DIR, NULL, MS_SHARED | MS_REC,
 			    NULL);
 	}
@@ -697,5 +783,54 @@ void sc_setup_user_mounts(struct sc_apparmor *apparmor, int snap_update_ns_fd,
 	// to slave mode, so we see changes from the parent namespace
 	// but don't propagate our own changes.
 	sc_do_mount("none", "/", NULL, MS_REC | MS_SLAVE, NULL);
+	sc_identity old = sc_set_effective_identity(sc_root_group_identity());
 	sc_call_snap_update_ns_as_user(snap_update_ns_fd, snap_name, apparmor);
+	(void)sc_set_effective_identity(old);
+}
+
+void sc_ensure_snap_dir_shared_mounts(void)
+{
+	const char *dirs[] = { SNAP_MOUNT_DIR, "/var/snap", NULL };
+	for (int i = 0; dirs[i] != NULL; i++) {
+		const char *dir = dirs[i];
+		if (!is_mounted_with_shared_option(dir)) {
+			/* Since this directory isn't yet shared (but it should be),
+			 * recursively bind mount it, then recursively share it so that
+			 * changes to the host are seen in the snap and vice-versa. This
+			 * allows us to fine-tune propagation events elsewhere for this new
+			 * mountpoint.
+			 *
+			 * Not using MS_SLAVE because it's too late for SNAP_MOUNT_DIR,
+			 * since snaps are already mounted, and it's not needed for
+			 * /var/snap.
+			 */
+			sc_do_mount(dir, dir, "none", MS_BIND | MS_REC, NULL);
+			sc_do_mount("none", dir, NULL, MS_REC | MS_SHARED,
+				    NULL);
+		}
+	}
+}
+
+void sc_setup_parallel_instance_classic_mounts(const char *snap_name,
+					       const char *snap_instance_name)
+{
+	char src[PATH_MAX] = { 0 };
+	char dst[PATH_MAX] = { 0 };
+
+	const char *dirs[] = { SNAP_MOUNT_DIR, "/var/snap", NULL };
+	for (int i = 0; dirs[i] != NULL; i++) {
+		const char *dir = dirs[i];
+		sc_do_mount("none", dir, NULL, MS_REC | MS_SLAVE, NULL);
+	}
+
+	/* Mount SNAP_MOUNT_DIR/<snap>_<key> on SNAP_MOUNT_DIR/<snap> */
+	sc_must_snprintf(src, sizeof src, "%s/%s", SNAP_MOUNT_DIR,
+			 snap_instance_name);
+	sc_must_snprintf(dst, sizeof dst, "%s/%s", SNAP_MOUNT_DIR, snap_name);
+	sc_do_mount(src, dst, "none", MS_BIND | MS_REC, NULL);
+
+	/* Mount /var/snap/<snap>_<key> on /var/snap/<snap> */
+	sc_must_snprintf(src, sizeof src, "/var/snap/%s", snap_instance_name);
+	sc_must_snprintf(dst, sizeof dst, "/var/snap/%s", snap_name);
+	sc_do_mount(src, dst, "none", MS_BIND | MS_REC, NULL);
 }
