@@ -92,12 +92,9 @@ network packet,
 /etc/netplan/{,**} rw,
 
 # Allow access to configuration files generated on the fly
-# from netplan and let NetworkManager store its DHCP leases
-# in the dhcp subdirectory so that console-conf can access
-# it.
-/run/NetworkManager/ w,
-/run/NetworkManager/{,**} r,
-/run/NetworkManager/dhcp/{,**} w,
+# from netplan and let NetworkManager store its configuration
+# in the same place.
+/run/NetworkManager/{,**} rw,
 
 # Needed by the ifupdown plugin to check which interfaces can
 # be managed an which not.
@@ -106,15 +103,15 @@ network packet,
 /etc/machine-id r,
 
 # Needed to use resolvconf from core
-/sbin/resolvconf ixr,
-/run/resolvconf/{,**} r,
+/{,usr/}sbin/resolvconf ixr,
+/run/resolvconf/{,**} rk,
 /run/resolvconf/** w,
 /etc/resolvconf/{,**} r,
-/lib/resolvconf/* ix,
+/{,usr/}lib/resolvconf/* ix,
 # NM peeks into ifupdown configuration
 /run/network/ifstate* r,
 # Required by resolvconf
-/bin/run-parts ixr,
+/{,usr/}bin/run-parts ixr,
 /etc/resolvconf/update.d/* ix,
 
 #include <abstractions/nameservice>
@@ -145,7 +142,7 @@ dbus (send)
      bus=system
      path="/org/freedesktop/resolve1"
      interface="org.freedesktop.resolve1.Manager"
-     member="SetLink{DNS,Domains}"
+     member="SetLink{DNS,MulticastDNS,Domains,LLMNR}"
      peer=(label=unconfined),
 
 dbus (send)
@@ -189,6 +186,13 @@ dbus (receive, send)
     interface=org.freedesktop.DBus.*
     peer=(label=unconfined),
 
+# Allow ObjectManager methods from and signals to unconfined clients.
+dbus (receive, send)
+    bus=system
+    path=/org/freedesktop
+    interface=org.freedesktop.DBus.ObjectManager
+    peer=(label=unconfined),
+
 # Allow access to hostname system service
 dbus (receive, send)
     bus=system
@@ -257,6 +261,15 @@ dbus (receive, send)
     path=/org/freedesktop/NetworkManager{,/**}
     peer=(label=###PLUG_SECURITY_TAGS###),
 
+# Later versions of NetworkManager implement org.freedesktop.DBus.ObjectManager
+# for clients to easily obtain all (and be alerted to added/removed) objects
+# from the service.
+dbus (receive, send)
+    bus=system
+    path=/org/freedesktop
+    interface=org.freedesktop.DBus.ObjectManager
+    peer=(label=###PLUG_SECURITY_TAGS###),
+
 # Explicitly deny ptrace to silence noisy denials. These denials happen when NM
 # tries to access /proc/<peer_pid>/stat.  What apparmor prevents is showing
 # internal process addresses that live in that file, but that has no adverse
@@ -276,6 +289,31 @@ dbus (receive, send)
     bus=system
     path=/org/freedesktop/NetworkManager{,/**}
     peer=(label=###SLOT_SECURITY_TAGS###),
+
+# NM implements org.freedesktop.DBus.ObjectManager too
+dbus (receive, send)
+    bus=system
+    path=/org/freedesktop
+    interface=org.freedesktop.DBus.ObjectManager
+    peer=(label=###SLOT_SECURITY_TAGS###),
+`
+
+const networkManagerConnectedPlugIntrospectionSnippet = `
+# Allow us to introspect the network-manager providing snap
+dbus (send)
+    bus=system
+    interface="org.freedesktop.DBus.Introspectable"
+    member="Introspect"
+    peer=(label=###SLOT_SECURITY_TAGS###),
+`
+
+const networkManagerConnectedSlotIntrospectionSnippet = `
+# Allow plugs to introspect us
+dbus (receive)
+    bus=system
+    interface="org.freedesktop.DBus.Introspectable"
+    member="Introspect"
+    peer=(label=###PLUG_SECURITY_TAGS###),
 `
 
 const networkManagerConnectedPlugSecComp = `
@@ -465,6 +503,11 @@ func (iface *networkManagerInterface) AppArmorConnectedPlug(spec *apparmor.Speci
 	}
 	snippet := strings.Replace(networkManagerConnectedPlugAppArmor, old, new, -1)
 	spec.AddSnippet(snippet)
+	if !release.OnClassic {
+		// See https://bugs.launchpad.net/snapd/+bug/1849291 for details.
+		snippet := strings.Replace(networkManagerConnectedPlugIntrospectionSnippet, old, new, -1)
+		spec.AddSnippet(snippet)
+	}
 	return nil
 }
 
@@ -473,6 +516,11 @@ func (iface *networkManagerInterface) AppArmorConnectedSlot(spec *apparmor.Speci
 	new := plugAppLabelExpr(plug)
 	snippet := strings.Replace(networkManagerConnectedSlotAppArmor, old, new, -1)
 	spec.AddSnippet(snippet)
+	if !release.OnClassic {
+		// See https://bugs.launchpad.net/snapd/+bug/1849291 for details.
+		snippet := strings.Replace(networkManagerConnectedSlotIntrospectionSnippet, old, new, -1)
+		spec.AddSnippet(snippet)
+	}
 	return nil
 }
 
