@@ -21,10 +21,12 @@ package overlord_test
 
 import (
 	"encoding/json"
+	"errors"
 	"fmt"
 	"io/ioutil"
 	"os"
 	"path/filepath"
+	"runtime"
 	"syscall"
 	"testing"
 	"time"
@@ -32,39 +34,67 @@ import (
 	. "gopkg.in/check.v1"
 	"gopkg.in/tomb.v2"
 
-	"github.com/snapcore/snapd/cmd"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/overlord"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/devicestate/devicestatetest"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/ifacestate"
 	"github.com/snapcore/snapd/overlord/patch"
+	"github.com/snapcore/snapd/overlord/restart"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
+	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snapdenv"
+	"github.com/snapcore/snapd/snapdtool"
 	"github.com/snapcore/snapd/store"
 	"github.com/snapcore/snapd/testutil"
+	"github.com/snapcore/snapd/timings"
 )
 
 func TestOverlord(t *testing.T) { TestingT(t) }
 
 type overlordSuite struct {
-	restoreBackends func()
+	testutil.BaseTest
 }
 
 var _ = Suite(&overlordSuite{})
 
-func (ovs *overlordSuite) SetUpTest(c *C) {
-	tmpdir := c.MkDir()
-	dirs.SetRootDir(tmpdir)
-	dirs.SnapStateFile = filepath.Join(tmpdir, "test.json")
-	snapstate.CanAutoRefresh = nil
-	ovs.restoreBackends = ifacestate.MockSecurityBackends(nil)
+type ticker struct {
+	tickerChannel chan time.Time
 }
 
-func (ovs *overlordSuite) TearDownTest(c *C) {
-	dirs.SetRootDir("/")
-	ovs.restoreBackends()
+func (w *ticker) tick(n int) {
+	for i := 0; i < n; i++ {
+		w.tickerChannel <- time.Now()
+	}
+}
+
+func fakePruneTicker() (w *ticker, restore func()) {
+	w = &ticker{
+		tickerChannel: make(chan time.Time),
+	}
+	restore = overlord.MockPruneTicker(func(t *time.Ticker) <-chan time.Time {
+		return w.tickerChannel
+	})
+	return w, restore
+}
+
+func (ovs *overlordSuite) SetUpTest(c *C) {
+	// temporary: skip due to timeouts on riscv64
+	if runtime.GOARCH == "riscv64" || os.Getenv("SNAPD_SKIP_SLOW_TESTS") != "" {
+		c.Skip("skipping slow test")
+	}
+
+	tmpdir := c.MkDir()
+	dirs.SetRootDir(tmpdir)
+	ovs.AddCleanup(func() { dirs.SetRootDir("") })
+	ovs.AddCleanup(osutil.MockMountInfo(""))
+
+	dirs.SnapStateFile = filepath.Join(tmpdir, "test.json")
+	snapstate.CanAutoRefresh = nil
+	ovs.AddCleanup(func() { ifacestate.MockSecurityBackends(nil) })
 }
 
 func (ovs *overlordSuite) TestNew(c *C) {
@@ -72,17 +102,19 @@ func (ovs *overlordSuite) TestNew(c *C) {
 	defer restore()
 
 	var configstateInitCalled bool
-	overlord.MockConfigstateInit(func(*hookstate.HookManager) {
+	overlord.MockConfigstateInit(func(*state.State, *hookstate.HookManager) error {
 		configstateInitCalled = true
+		return nil
 	})
 
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 	c.Check(o, NotNil)
 
 	c.Check(o.StateEngine(), NotNil)
 	c.Check(o.TaskRunner(), NotNil)
 	c.Check(o.SnapManager(), NotNil)
+	c.Check(o.ServiceManager(), NotNil)
 	c.Check(o.AssertManager(), NotNil)
 	c.Check(o.InterfaceManager(), NotNil)
 	c.Check(o.HookManager(), NotNil)
@@ -117,7 +149,7 @@ func (ovs *overlordSuite) TestNew(c *C) {
 func (ovs *overlordSuite) TestNewStore(c *C) {
 	// this is a shallow test, the deep testing happens in the
 	// remodeling tests in managers_test.go
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 
 	devBE := o.DeviceManager().StoreContextBackend()
@@ -128,11 +160,17 @@ func (ovs *overlordSuite) TestNewStore(c *C) {
 }
 
 func (ovs *overlordSuite) TestNewWithGoodState(c *C) {
-	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":%d,"patch-sublevel":%d,"patch-sublevel-last-version":%q,"some":"data","refresh-privacy-key":"0123456789ABCDEF"},"changes":null,"tasks":null,"last-change-id":0,"last-task-id":0,"last-lane-id":0}`, patch.Level, patch.Sublevel, cmd.Version))
+	// ensure we don't write state load timing in the state on really
+	// slow architectures (e.g. risc-v)
+	oldDurationThreshold := timings.DurationThreshold
+	timings.DurationThreshold = time.Second * 30
+	defer func() { timings.DurationThreshold = oldDurationThreshold }()
+
+	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":%d,"patch-sublevel":%d,"patch-sublevel-last-version":%q,"some":"data","refresh-privacy-key":"0123456789ABCDEF"},"changes":null,"tasks":null,"last-change-id":0,"last-task-id":0,"last-lane-id":0}`, patch.Level, patch.Sublevel, snapdtool.Version))
 	err := ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
 	c.Assert(err, IsNil)
 
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 
 	state := o.State()
@@ -160,7 +198,7 @@ func (ovs *overlordSuite) TestNewWithStateSnapmgrUpdate(c *C) {
 	err := ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
 	c.Assert(err, IsNil)
 
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 
 	state := o.State()
@@ -178,7 +216,7 @@ func (ovs *overlordSuite) TestNewWithInvalidState(c *C) {
 	err := ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
 	c.Assert(err, IsNil)
 
-	_, err = overlord.New()
+	_, err = overlord.New(nil)
 	c.Assert(err, ErrorMatches, "cannot read state: EOF")
 }
 
@@ -193,11 +231,11 @@ func (ovs *overlordSuite) TestNewWithPatches(c *C) {
 	}
 	patch.Mock(1, 1, map[int][]patch.PatchFunc{1: {p, sp}})
 
-	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":0, "patch-sublevel":0}}`))
+	fakeState := []byte(`{"data":{"patch-level":0, "patch-sublevel":0}}`)
 	err := ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
 	c.Assert(err, IsNil)
 
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 
 	state := o.State()
@@ -223,11 +261,34 @@ func (ovs *overlordSuite) TestNewWithPatches(c *C) {
 	c.Check(b, Equals, true)
 }
 
+func (ovs *overlordSuite) TestNewFailedConfigstate(c *C) {
+	restore := patch.Mock(42, 2, nil)
+	defer restore()
+
+	var configstateInitCalled bool
+	restore = overlord.MockConfigstateInit(func(*state.State, *hookstate.HookManager) error {
+		configstateInitCalled = true
+		return fmt.Errorf("bad bad")
+	})
+	defer restore()
+
+	o, err := overlord.New(nil)
+	c.Assert(err, ErrorMatches, "bad bad")
+	c.Check(o, IsNil)
+	c.Check(configstateInitCalled, Equals, true)
+}
+
 type witnessManager struct {
 	state          *state.State
 	expectedEnsure int
 	ensureCalled   chan struct{}
 	ensureCallback func(s *state.State) error
+	startedUp      int
+}
+
+func (wm *witnessManager) StartUp() error {
+	wm.startedUp++
+	return nil
 }
 
 func (wm *witnessManager) Ensure() error {
@@ -255,13 +316,16 @@ func markSeeded(o *overlord.Overlord) {
 }
 
 func (ovs *overlordSuite) TestTrivialRunAndStop(c *C) {
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 
 	markSeeded(o)
 	// make sure we don't try to talk to the store
 	snapstate.CanAutoRefresh = nil
 
+	err = o.StartUp()
+	c.Assert(err, IsNil)
+
 	o.Loop()
 
 	err = o.Stop()
@@ -269,7 +333,7 @@ func (ovs *overlordSuite) TestTrivialRunAndStop(c *C) {
 }
 
 func (ovs *overlordSuite) TestUnknownTasks(c *C) {
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 	o.InterfaceManager().DisableUDevMonitor()
 
@@ -305,6 +369,9 @@ func (ovs *overlordSuite) TestEnsureLoopRunAndStop(c *C) {
 	}
 	o.AddManager(witness)
 
+	err := o.StartUp()
+	c.Assert(err, IsNil)
+
 	o.Loop()
 	defer o.Stop()
 
@@ -316,8 +383,10 @@ func (ovs *overlordSuite) TestEnsureLoopRunAndStop(c *C) {
 	}
 	c.Check(time.Since(t0) >= 10*time.Millisecond, Equals, true)
 
-	err := o.Stop()
+	err = o.Stop()
 	c.Assert(err, IsNil)
+
+	c.Check(witness.startedUp, Equals, 1)
 }
 
 func (ovs *overlordSuite) TestEnsureLoopMediatedEnsureBeforeImmediate(c *C) {
@@ -338,6 +407,8 @@ func (ovs *overlordSuite) TestEnsureLoopMediatedEnsureBeforeImmediate(c *C) {
 	}
 	o.AddManager(witness)
 
+	c.Assert(o.StartUp(), IsNil)
+
 	o.Loop()
 	defer o.Stop()
 
@@ -366,6 +437,8 @@ func (ovs *overlordSuite) TestEnsureLoopMediatedEnsureBefore(c *C) {
 	}
 	o.AddManager(witness)
 
+	c.Assert(o.StartUp(), IsNil)
+
 	o.Loop()
 	defer o.Stop()
 
@@ -395,6 +468,8 @@ func (ovs *overlordSuite) TestEnsureBeforeSleepy(c *C) {
 	}
 	o.AddManager(witness)
 
+	c.Assert(o.StartUp(), IsNil)
+
 	o.Loop()
 	defer o.Stop()
 
@@ -424,6 +499,8 @@ func (ovs *overlordSuite) TestEnsureBeforeLater(c *C) {
 	}
 	o.AddManager(witness)
 
+	c.Assert(o.StartUp(), IsNil)
+
 	o.Loop()
 	defer o.Stop()
 
@@ -453,6 +530,8 @@ func (ovs *overlordSuite) TestEnsureLoopMediatedEnsureBeforeOutsideEnsure(c *C)
 	}
 	o.AddManager(witness)
 
+	c.Assert(o.StartUp(), IsNil)
+
 	o.Loop()
 	defer o.Stop()
 
@@ -509,6 +588,8 @@ func (ovs *overlordSuite) TestEnsureLoopPrune(c *C) {
 	}
 	o.AddManager(witness)
 
+	c.Assert(o.StartUp(), IsNil)
+
 	o.Loop()
 
 	select {
@@ -530,7 +611,7 @@ func (ovs *overlordSuite) TestEnsureLoopPrune(c *C) {
 }
 
 func (ovs *overlordSuite) TestEnsureLoopPruneRunsMultipleTimes(c *C) {
-	restoreIntv := overlord.MockPruneInterval(100*time.Millisecond, 1000*time.Millisecond, 1*time.Hour)
+	restoreIntv := overlord.MockPruneInterval(100*time.Millisecond, 5*time.Millisecond, 1*time.Hour)
 	defer restoreIntv()
 	o := overlord.Mock()
 
@@ -548,20 +629,26 @@ func (ovs *overlordSuite) TestEnsureLoopPruneRunsMultipleTimes(c *C) {
 	c.Check(st.Changes(), HasLen, 2)
 	st.Unlock()
 
+	w, restoreTicker := fakePruneTicker()
+	defer restoreTicker()
+
 	// start the loop that runs the prune ticker
 	o.Loop()
 
-	// ensure the first change is pruned
-	time.Sleep(1500 * time.Millisecond)
+	// this needs to be more than pruneWait=5ms mocked above
+	time.Sleep(10 * time.Millisecond)
+	w.tick(2)
+
 	st.Lock()
 	c.Check(st.Changes(), HasLen, 1)
-	st.Unlock()
-
-	// ensure the second is also purged after it is ready
-	st.Lock()
 	chg2.SetStatus(state.DoneStatus)
 	st.Unlock()
-	time.Sleep(1500 * time.Millisecond)
+
+	// this needs to be more than pruneWait=5ms mocked above
+	time.Sleep(10 * time.Millisecond)
+	// tick twice for extra Ensure
+	w.tick(2)
+
 	st.Lock()
 	c.Check(st.Changes(), HasLen, 0)
 	st.Unlock()
@@ -571,11 +658,197 @@ func (ovs *overlordSuite) TestEnsureLoopPruneRunsMultipleTimes(c *C) {
 	c.Assert(err, IsNil)
 }
 
+func (ovs *overlordSuite) TestOverlordStartUpSetsStartOfOperation(c *C) {
+	restoreIntv := overlord.MockPruneInterval(100*time.Millisecond, 1000*time.Millisecond, 1*time.Hour)
+	defer restoreIntv()
+
+	// use real overlord, we need device manager to be there
+	o, err := overlord.New(nil)
+	c.Assert(err, IsNil)
+
+	st := o.State()
+	st.Lock()
+	defer st.Unlock()
+
+	// sanity check, not set
+	var opTime time.Time
+	c.Assert(st.Get("start-of-operation-time", &opTime), Equals, state.ErrNoState)
+	st.Unlock()
+
+	c.Assert(o.StartUp(), IsNil)
+
+	st.Lock()
+	c.Assert(st.Get("start-of-operation-time", &opTime), IsNil)
+}
+
+func (ovs *overlordSuite) TestEnsureLoopPruneDoesntAbortShortlyAfterStartOfOperation(c *C) {
+	w, restoreTicker := fakePruneTicker()
+	defer restoreTicker()
+
+	// use real overlord, we need device manager to be there
+	o, err := overlord.New(nil)
+	c.Assert(err, IsNil)
+
+	markSeeded(o)
+
+	// avoid immediate transition to Done due to unknown kind
+	o.TaskRunner().AddHandler("bar", func(t *state.Task, _ *tomb.Tomb) error {
+		return &state.Retry{}
+	}, nil)
+
+	st := o.State()
+	st.Lock()
+
+	// start of operation time is 50min ago, this is less then abort limit
+	opTime := time.Now().Add(-50 * time.Minute)
+	st.Set("start-of-operation-time", opTime)
+
+	// spawn time one month ago
+	spawnTime := time.Now().AddDate(0, -1, 0)
+	restoreTimeNow := state.MockTime(spawnTime)
+
+	t := st.NewTask("bar", "...")
+	chg := st.NewChange("other-change", "...")
+	chg.AddTask(t)
+
+	restoreTimeNow()
+
+	// sanity
+	c.Check(st.Changes(), HasLen, 1)
+
+	st.Unlock()
+	c.Assert(o.StartUp(), IsNil)
+
+	// start the loop that runs the prune ticker
+	o.Loop()
+	w.tick(2)
+
+	c.Assert(o.Stop(), IsNil)
+
+	st.Lock()
+	defer st.Unlock()
+	c.Assert(st.Changes(), HasLen, 1)
+	c.Check(chg.Status(), Equals, state.DoingStatus)
+}
+
+func (ovs *overlordSuite) TestEnsureLoopPruneAbortsOld(c *C) {
+	// Ensure interval is not relevant for this test
+	restoreEnsureIntv := overlord.MockEnsureInterval(10 * time.Hour)
+	defer restoreEnsureIntv()
+
+	w, restoreTicker := fakePruneTicker()
+	defer restoreTicker()
+
+	// use real overlord, we need device manager to be there
+	o, err := overlord.New(nil)
+	c.Assert(err, IsNil)
+
+	// avoid immediate transition to Done due to having unknown kind
+	o.TaskRunner().AddHandler("bar", func(t *state.Task, _ *tomb.Tomb) error {
+		return &state.Retry{}
+	}, nil)
+
+	markSeeded(o)
+
+	st := o.State()
+	st.Lock()
+
+	// start of operation time is a year ago
+	opTime := time.Now().AddDate(-1, 0, 0)
+	st.Set("start-of-operation-time", opTime)
+
+	st.Unlock()
+	c.Assert(o.StartUp(), IsNil)
+	st.Lock()
+
+	// spawn time one month ago
+	spawnTime := time.Now().AddDate(0, -1, 0)
+	restoreTimeNow := state.MockTime(spawnTime)
+	t := st.NewTask("bar", "...")
+	chg := st.NewChange("other-change", "...")
+	chg.AddTask(t)
+
+	restoreTimeNow()
+
+	// sanity
+	c.Check(st.Changes(), HasLen, 1)
+	st.Unlock()
+
+	// start the loop that runs the prune ticker
+	o.Loop()
+	w.tick(2)
+
+	c.Assert(o.Stop(), IsNil)
+
+	st.Lock()
+	defer st.Unlock()
+
+	// sanity
+	op, err := o.DeviceManager().StartOfOperationTime()
+	c.Assert(err, IsNil)
+	c.Check(op.Equal(opTime), Equals, true)
+
+	c.Assert(st.Changes(), HasLen, 1)
+	// change was aborted
+	c.Check(chg.Status(), Equals, state.HoldStatus)
+}
+
+func (ovs *overlordSuite) TestEnsureLoopNoPruneWhenPreseed(c *C) {
+	w, restoreTicker := fakePruneTicker()
+	defer restoreTicker()
+
+	restore := snapdenv.MockPreseeding(true)
+	defer restore()
+
+	restorePreseedExitWithErr := overlord.MockPreseedExitWithError(func(err error) {})
+	defer restorePreseedExitWithErr()
+
+	o, err := overlord.New(nil)
+	c.Assert(err, IsNil)
+	markSeeded(o)
+
+	// avoid immediate transition to Done due to unknown kind
+	o.TaskRunner().AddHandler("bar", func(t *state.Task, _ *tomb.Tomb) error {
+		return &state.Retry{}
+	}, nil)
+
+	// sanity
+	_, err = o.DeviceManager().StartOfOperationTime()
+	c.Assert(err, ErrorMatches, `internal error: unexpected call to StartOfOperationTime in preseed mode`)
+
+	c.Assert(o.StartUp(), IsNil)
+
+	st := o.State()
+	st.Lock()
+
+	// spawn time one month ago
+	spawnTime := time.Now().AddDate(0, -1, 0)
+	restoreTimeNow := state.MockTime(spawnTime)
+	t := st.NewTask("bar", "...")
+	chg := st.NewChange("change", "...")
+	chg.AddTask(t)
+	restoreTimeNow()
+
+	st.Unlock()
+	// start the loop that runs the prune ticker
+	o.Loop()
+	w.tick(2)
+
+	c.Assert(o.Stop(), IsNil)
+
+	st.Lock()
+	defer st.Unlock()
+
+	var opTime time.Time
+	c.Assert(st.Get("start-of-operation-time", &opTime), Equals, state.ErrNoState)
+	c.Check(chg.Status(), Equals, state.DoingStatus)
+}
+
 func (ovs *overlordSuite) TestCheckpoint(c *C) {
 	oldUmask := syscall.Umask(0)
 	defer syscall.Umask(oldUmask)
 
-	o, err := overlord.New()
+	o, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 
 	s := o.State()
@@ -594,7 +867,7 @@ type sampleManager struct {
 	ensureCallback func()
 }
 
-func newSampleManager(s *state.State, runner *state.TaskRunner) *sampleManager {
+func newSampleManager(runner *state.TaskRunner) *sampleManager {
 	sm := &sampleManager{}
 
 	runner.AddHandler("runMgr1", func(t *state.Task, _ *tomb.Tomb) error {
@@ -656,7 +929,7 @@ func (ovs *overlordSuite) TestTrivialSettle(c *C) {
 	o := overlord.Mock()
 
 	s := o.State()
-	sm1 := newSampleManager(s, o.TaskRunner())
+	sm1 := newSampleManager(o.TaskRunner())
 	o.AddManager(sm1)
 	o.AddManager(o.TaskRunner())
 
@@ -685,7 +958,7 @@ func (ovs *overlordSuite) TestSettleNotConverging(c *C) {
 	o := overlord.Mock()
 
 	s := o.State()
-	sm1 := newSampleManager(s, o.TaskRunner())
+	sm1 := newSampleManager(o.TaskRunner())
 	o.AddManager(sm1)
 	o.AddManager(o.TaskRunner())
 
@@ -712,7 +985,7 @@ func (ovs *overlordSuite) TestSettleChain(c *C) {
 	o := overlord.Mock()
 
 	s := o.State()
-	sm1 := newSampleManager(s, o.TaskRunner())
+	sm1 := newSampleManager(o.TaskRunner())
 	o.AddManager(sm1)
 	o.AddManager(o.TaskRunner())
 
@@ -746,7 +1019,7 @@ func (ovs *overlordSuite) TestSettleChainWCleanup(c *C) {
 	o := overlord.Mock()
 
 	s := o.State()
-	sm1 := newSampleManager(s, o.TaskRunner())
+	sm1 := newSampleManager(o.TaskRunner())
 	o.AddManager(sm1)
 	o.AddManager(o.TaskRunner())
 
@@ -783,7 +1056,7 @@ func (ovs *overlordSuite) TestSettleExplicitEnsureBefore(c *C) {
 	o := overlord.Mock()
 
 	s := o.State()
-	sm1 := newSampleManager(s, o.TaskRunner())
+	sm1 := newSampleManager(o.TaskRunner())
 	sm1.ensureCallback = func() {
 		s.Lock()
 		defer s.Unlock()
@@ -815,26 +1088,165 @@ func (ovs *overlordSuite) TestSettleExplicitEnsureBefore(c *C) {
 	c.Check(v, Equals, 2)
 }
 
-func (ovs *overlordSuite) TestRequestRestartNoHandler(c *C) {
-	o, err := overlord.New()
+func (ovs *overlordSuite) TestEnsureErrorWhenPreseeding(c *C) {
+	restore := snapdenv.MockPreseeding(true)
+	defer restore()
+
+	restoreIntv := overlord.MockEnsureInterval(1 * time.Millisecond)
+	defer restoreIntv()
+
+	var errorCallbackError error
+	restoreExitWithErr := overlord.MockPreseedExitWithError(func(err error) {
+		errorCallbackError = err
+	})
+	defer restoreExitWithErr()
+
+	o, err := overlord.New(nil)
+	c.Assert(err, IsNil)
+	markSeeded(o)
+
+	err = o.StartUp()
 	c.Assert(err, IsNil)
 
-	o.State().RequestRestart(state.RestartDaemon)
+	o.TaskRunner().AddHandler("bar", func(t *state.Task, _ *tomb.Tomb) error {
+		return fmt.Errorf("bar error")
+	}, nil)
+
+	s := o.State()
+	s.Lock()
+	defer s.Unlock()
+
+	chg := s.NewChange("chg", "...")
+	t := s.NewTask("bar", "...")
+	chg.AddTask(t)
+
+	s.Unlock()
+	o.Loop()
+	time.Sleep(1 * time.Second)
+	o.Stop()
+
+	s.Lock()
+	c.Check(t.Status(), Equals, state.ErrorStatus)
+	c.Check(errorCallbackError, ErrorMatches, "bar error")
+}
+
+func (ovs *overlordSuite) TestRequestRestartNoHandler(c *C) {
+	o, err := overlord.New(nil)
+	c.Assert(err, IsNil)
+
+	st := o.State()
+	st.Lock()
+	defer st.Unlock()
+
+	restart.Request(st, restart.RestartDaemon)
+}
+
+type testRestartHandler struct {
+	restartRequested  restart.RestartType
+	rebootState       string
+	rebootVerifiedErr error
+}
+
+func (rb *testRestartHandler) HandleRestart(t restart.RestartType) {
+	rb.restartRequested = t
+}
+
+func (rb *testRestartHandler) RebootAsExpected(_ *state.State) error {
+	rb.rebootState = "as-expected"
+	return rb.rebootVerifiedErr
+}
+
+func (rb *testRestartHandler) RebootDidNotHappen(_ *state.State) error {
+	rb.rebootState = "did-not-happen"
+	return rb.rebootVerifiedErr
 }
 
 func (ovs *overlordSuite) TestRequestRestartHandler(c *C) {
-	o, err := overlord.New()
+	rb := &testRestartHandler{}
+
+	o, err := overlord.New(rb)
 	c.Assert(err, IsNil)
 
-	restartRequested := false
+	st := o.State()
+	st.Lock()
+	defer st.Unlock()
 
-	o.SetRestartHandler(func(t state.RestartType) {
-		restartRequested = true
-	})
+	restart.Request(st, restart.RestartDaemon)
 
-	o.State().RequestRestart(state.RestartDaemon)
+	c.Check(rb.restartRequested, Equals, restart.RestartDaemon)
+}
 
-	c.Check(restartRequested, Equals, true)
+func (ovs *overlordSuite) TestVerifyRebootNoPendingReboot(c *C) {
+	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":%d,"patch-sublevel":%d,"some":"data","refresh-privacy-key":"0123456789ABCDEF"},"changes":null,"tasks":null,"last-change-id":0,"last-task-id":0,"last-lane-id":0}`, patch.Level, patch.Sublevel))
+	err := ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
+	c.Assert(err, IsNil)
+
+	rb := &testRestartHandler{}
+
+	_, err = overlord.New(rb)
+	c.Assert(err, IsNil)
+
+	c.Check(rb.rebootState, Equals, "as-expected")
+}
+
+func (ovs *overlordSuite) TestVerifyRebootOK(c *C) {
+	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":%d,"patch-sublevel":%d,"some":"data","refresh-privacy-key":"0123456789ABCDEF","system-restart-from-boot-id":%q},"changes":null,"tasks":null,"last-change-id":0,"last-task-id":0,"last-lane-id":0}`, patch.Level, patch.Sublevel, "boot-id-prev"))
+	err := ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
+	c.Assert(err, IsNil)
+
+	rb := &testRestartHandler{}
+
+	_, err = overlord.New(rb)
+	c.Assert(err, IsNil)
+
+	c.Check(rb.rebootState, Equals, "as-expected")
+}
+
+func (ovs *overlordSuite) TestVerifyRebootOKButError(c *C) {
+	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":%d,"patch-sublevel":%d,"some":"data","refresh-privacy-key":"0123456789ABCDEF","system-restart-from-boot-id":%q},"changes":null,"tasks":null,"last-change-id":0,"last-task-id":0,"last-lane-id":0}`, patch.Level, patch.Sublevel, "boot-id-prev"))
+	err := ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
+	c.Assert(err, IsNil)
+
+	e := errors.New("boom")
+	rb := &testRestartHandler{rebootVerifiedErr: e}
+
+	_, err = overlord.New(rb)
+	c.Assert(err, Equals, e)
+
+	c.Check(rb.rebootState, Equals, "as-expected")
+}
+
+func (ovs *overlordSuite) TestVerifyRebootDidNotHappen(c *C) {
+	curBootID, err := osutil.BootID()
+	c.Assert(err, IsNil)
+
+	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":%d,"patch-sublevel":%d,"some":"data","refresh-privacy-key":"0123456789ABCDEF","system-restart-from-boot-id":%q},"changes":null,"tasks":null,"last-change-id":0,"last-task-id":0,"last-lane-id":0}`, patch.Level, patch.Sublevel, curBootID))
+	err = ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
+	c.Assert(err, IsNil)
+
+	rb := &testRestartHandler{}
+
+	_, err = overlord.New(rb)
+	c.Assert(err, IsNil)
+
+	c.Check(rb.rebootState, Equals, "did-not-happen")
+}
+
+func (ovs *overlordSuite) TestVerifyRebootDidNotHappenError(c *C) {
+	curBootID, err := osutil.BootID()
+	c.Assert(err, IsNil)
+
+	fakeState := []byte(fmt.Sprintf(`{"data":{"patch-level":%d,"patch-sublevel":%d,"some":"data","refresh-privacy-key":"0123456789ABCDEF","system-restart-from-boot-id":%q},"changes":null,"tasks":null,"last-change-id":0,"last-task-id":0,"last-lane-id":0}`, patch.Level, patch.Sublevel, curBootID))
+	err = ioutil.WriteFile(dirs.SnapStateFile, fakeState, 0600)
+	c.Assert(err, IsNil)
+
+	e := errors.New("boom")
+	rb := &testRestartHandler{rebootVerifiedErr: e}
+
+	_, err = overlord.New(rb)
+	c.Assert(err, Equals, e)
+
+	c.Check(rb.rebootState, Equals, "did-not-happen")
 }
 
 func (ovs *overlordSuite) TestOverlordCanStandby(c *C) {
@@ -848,6 +1260,8 @@ func (ovs *overlordSuite) TestOverlordCanStandby(c *C) {
 	}
 	o.AddManager(witness)
 
+	c.Assert(o.StartUp(), IsNil)
+
 	// can only standby after loop ran once
 	c.Assert(o.CanStandby(), Equals, false)
 
@@ -862,3 +1276,42 @@ func (ovs *overlordSuite) TestOverlordCanStandby(c *C) {
 
 	c.Assert(o.CanStandby(), Equals, true)
 }
+
+func (ovs *overlordSuite) TestStartupTimeout(c *C) {
+	o, err := overlord.New(nil)
+	c.Assert(err, IsNil)
+
+	to, _, err := o.StartupTimeout()
+	c.Assert(err, IsNil)
+	c.Check(to, Equals, 30*time.Second)
+
+	st := o.State()
+	st.Lock()
+	defer st.Unlock()
+
+	// have two snaps
+	snapstate.Set(st, "core18", &snapstate.SnapState{
+		Active: true,
+		Sequence: []*snap.SideInfo{
+			{RealName: "core18", Revision: snap.R(1)},
+		},
+		Current:  snap.R(1),
+		SnapType: "base",
+	})
+	snapstate.Set(st, "foo", &snapstate.SnapState{
+		Active: true,
+		Sequence: []*snap.SideInfo{
+			{RealName: "foo", Revision: snap.R(1)},
+		},
+		Current:  snap.R(1),
+		SnapType: "app",
+	})
+
+	st.Unlock()
+	to, reasoning, err := o.StartupTimeout()
+	st.Lock()
+	c.Assert(err, IsNil)
+
+	c.Check(to, Equals, (30+5+5)*time.Second)
+	c.Check(reasoning, Equals, "pessimistic estimate of 30s plus 5s per snap")
+}
