@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2021 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -22,6 +22,7 @@ package snap
 import (
 	"errors"
 	"fmt"
+	"net/url"
 	"os"
 	"path/filepath"
 	"regexp"
@@ -30,6 +31,7 @@ import (
 	"strings"
 	"unicode/utf8"
 
+	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/snap/naming"
 	"github.com/snapcore/snapd/spdx"
 	"github.com/snapcore/snapd/strutil"
@@ -37,9 +39,6 @@ import (
 	"github.com/snapcore/snapd/timeutil"
 )
 
-// The fixed length of valid snap IDs.
-const validSnapIDLength = 32
-
 // ValidateInstanceName checks if a string can be used as a snap instance name.
 func ValidateInstanceName(instanceName string) error {
 	return naming.ValidateInstance(instanceName)
@@ -50,6 +49,25 @@ func ValidateName(name string) error {
 	return naming.ValidateSnap(name)
 }
 
+// ValidateDesktopPrefix checks if a string can be used as a desktop file
+// prefix. A desktop prefix should be of the form 'snapname' or
+// 'snapname+instance'.
+func ValidateDesktopPrefix(prefix string) bool {
+	tokens := strings.Split(prefix, "+")
+	if len(tokens) == 0 || len(tokens) > 2 {
+		return false
+	}
+	if err := ValidateName(tokens[0]); err != nil {
+		return false
+	}
+	if len(tokens) == 2 {
+		if err := ValidateInstanceName(tokens[1]); err != nil {
+			return false
+		}
+	}
+	return true
+}
+
 // ValidatePlugName checks if a string can be used as a slot name.
 //
 // Slot names and plug names within one snap must have unique names.
@@ -200,9 +218,19 @@ func validateSocketAddrPath(socket *SocketInfo, fieldName string, path string) e
 		return fmt.Errorf("invalid %q: %q should be written as %q", fieldName, path, clean)
 	}
 
-	if !(strings.HasPrefix(path, "$SNAP_DATA/") || strings.HasPrefix(path, "$SNAP_COMMON/") || strings.HasPrefix(path, "$XDG_RUNTIME_DIR/")) {
-		return fmt.Errorf(
-			"invalid %q: must have a prefix of $SNAP_DATA, $SNAP_COMMON or $XDG_RUNTIME_DIR", fieldName)
+	switch socket.App.DaemonScope {
+	case SystemDaemon:
+		if !(strings.HasPrefix(path, "$SNAP_DATA/") || strings.HasPrefix(path, "$SNAP_COMMON/") || strings.HasPrefix(path, "$XDG_RUNTIME_DIR/")) {
+			return fmt.Errorf(
+				"invalid %q: system daemon sockets must have a prefix of $SNAP_DATA, $SNAP_COMMON or $XDG_RUNTIME_DIR", fieldName)
+		}
+	case UserDaemon:
+		if !(strings.HasPrefix(path, "$SNAP_USER_DATA/") || strings.HasPrefix(path, "$SNAP_USER_COMMON/") || strings.HasPrefix(path, "$XDG_RUNTIME_DIR/")) {
+			return fmt.Errorf(
+				"invalid %q: user daemon sockets must have a prefix of $SNAP_USER_DATA, $SNAP_USER_COMMON, or $XDG_RUNTIME_DIR", fieldName)
+		}
+	default:
+		return fmt.Errorf("invalid %q: cannot validate sockets for daemon-scope %q", fieldName, socket.App.DaemonScope)
 	}
 
 	return nil
@@ -221,17 +249,17 @@ func validateSocketAddrAbstract(socket *SocketInfo, fieldName string, path strin
 func validateSocketAddrNet(socket *SocketInfo, fieldName string, address string) error {
 	lastIndex := strings.LastIndex(address, ":")
 	if lastIndex >= 0 {
-		if err := validateSocketAddrNetHost(socket, fieldName, address[:lastIndex]); err != nil {
+		if err := validateSocketAddrNetHost(fieldName, address[:lastIndex]); err != nil {
 			return err
 		}
-		return validateSocketAddrNetPort(socket, fieldName, address[lastIndex+1:])
+		return validateSocketAddrNetPort(fieldName, address[lastIndex+1:])
 	}
 
 	// Address only contains a port
-	return validateSocketAddrNetPort(socket, fieldName, address)
+	return validateSocketAddrNetPort(fieldName, address)
 }
 
-func validateSocketAddrNetHost(socket *SocketInfo, fieldName string, address string) error {
+func validateSocketAddrNetHost(fieldName string, address string) error {
 	validAddresses := []string{"127.0.0.1", "[::1]", "[::]"}
 	for _, valid := range validAddresses {
 		if address == valid {
@@ -242,7 +270,7 @@ func validateSocketAddrNetHost(socket *SocketInfo, fieldName string, address str
 	return fmt.Errorf("invalid %q address %q, must be one of: %s", fieldName, address, strings.Join(validAddresses, ", "))
 }
 
-func validateSocketAddrNetPort(socket *SocketInfo, fieldName string, port string) error {
+func validateSocketAddrNetPort(fieldName string, port string) error {
 	var val uint64
 	var err error
 	retErr := fmt.Errorf("invalid %q port number %q", fieldName, port)
@@ -346,6 +374,16 @@ func Validate(info *Info) error {
 		return err
 	}
 
+	// Ensure system usernames are valid
+	if err := ValidateSystemUsernames(info); err != nil {
+		return err
+	}
+
+	// Ensure links are valid
+	if err := ValidateLinks(info.Links()); err != nil {
+		return err
+	}
+
 	// ensure that common-id(s) are unique
 	if err := ValidateCommonIDs(info); err != nil {
 		return err
@@ -357,9 +395,9 @@ func Validate(info *Info) error {
 // ValidateBase validates the base field.
 func ValidateBase(info *Info) error {
 	// validate that bases do not have base fields
-	if info.GetType() == TypeOS || info.GetType() == TypeBase {
+	if info.Type() == TypeOS || info.Type() == TypeBase {
 		if info.Base != "" && info.Base != "none" {
-			return fmt.Errorf(`cannot have "base" field on %q snap %q`, info.GetType(), info.InstanceName())
+			return fmt.Errorf(`cannot have "base" field on %q snap %q`, info.Type(), info.InstanceName())
 		}
 	}
 
@@ -387,6 +425,38 @@ func ValidateLayoutAll(info *Info) error {
 	}
 	sort.Strings(paths)
 
+	// Validate that each source path is not a new top-level directory
+	for _, layout := range info.Layout {
+		cleanPathSrc := info.ExpandSnapVariables(filepath.Clean(layout.Path))
+		elems := strings.SplitN(cleanPathSrc, string(os.PathSeparator), 3)
+		switch len(elems) {
+		// len(1) is either relative path or empty string, will be validated
+		// elsewhere
+		case 2, 3:
+			// if the first string is the empty string, then we have a top-level
+			// directory to check
+			if elems[0] != "" {
+				// not the empty string which means this was a relative
+				// specification, i.e. usr/src/doc
+				return fmt.Errorf("layout %q is a relative filename", layout.Path)
+			}
+			if elems[1] != "" {
+				// verify that the top-level directory is a supported one
+				// we can't create new top-level directories because that would
+				// require creating a mimic on top of "/" which we don't
+				// currently support
+				switch elems[1] {
+				// this list was produced by taking all of the top level
+				// directories in the core snap and removing the explicitly
+				// denied top-level directories
+				case "bin", "etc", "lib", "lib64", "meta", "mnt", "opt", "root", "sbin", "snap", "srv", "usr", "var", "writable":
+				default:
+					return fmt.Errorf("layout %q defines a new top-level directory %q", layout.Path, "/"+elems[1])
+				}
+			}
+		}
+	}
+
 	// Validate that each source path is used consistently as a file or as a directory.
 	sourceKindMap := make(map[string]string)
 	for _, path := range paths {
@@ -413,6 +483,16 @@ func ValidateLayoutAll(info *Info) error {
 		}
 	}
 
+	// Validate that layout are not attempting to define elements that normally
+	// come from other snaps. This is separate from the ValidateLayout below to
+	// simplify argument passing.
+	thisSnapMntDir := filepath.Join("/snap/", info.SnapName())
+	for _, path := range paths {
+		if strings.HasPrefix(path, "/snap/") && !strings.HasPrefix(path, thisSnapMntDir) {
+			return fmt.Errorf("layout %q defines a layout in space belonging to another snap", path)
+		}
+	}
+
 	// Validate each layout item and collect resulting constraints.
 	constraints := make([]LayoutConstraint, 0, len(info.Layout))
 	for _, path := range paths {
@@ -497,6 +577,10 @@ func validateAppOrderNames(app *AppInfo, dependencies []string) error {
 		if !other.IsService() {
 			return fmt.Errorf("before/after references a non-service application %q", dep)
 		}
+
+		if app.DaemonScope != other.DaemonScope {
+			return fmt.Errorf("before/after references service with different daemon-scope %q", dep)
+		}
 	}
 	return nil
 }
@@ -565,6 +649,43 @@ func validateAppRestart(app *AppInfo) error {
 	return nil
 }
 
+func validateAppActivatesOn(app *AppInfo) error {
+	if len(app.ActivatesOn) == 0 {
+		return nil
+	}
+
+	if !app.IsService() {
+		return errors.New("activates-on is only applicable to services")
+	}
+
+	for _, slot := range app.ActivatesOn {
+		// ActivatesOn slots must use the "dbus" interface
+		if slot.Interface != "dbus" {
+			return fmt.Errorf("invalid activates-on value %q: slot does not use dbus interface", slot.Name)
+		}
+
+		// D-Bus slots must match the daemon scope
+		bus := slot.Attrs["bus"]
+		if app.DaemonScope == SystemDaemon && bus != "system" || app.DaemonScope == UserDaemon && bus != "session" {
+			return fmt.Errorf("invalid activates-on value %q: bus %q does not match daemon-scope %q", slot.Name, bus, app.DaemonScope)
+		}
+
+		// Slots must only be activatable on a single app
+		for _, otherApp := range slot.Apps {
+			if otherApp == app {
+				continue
+			}
+			for _, otherSlot := range otherApp.ActivatesOn {
+				if otherSlot == slot {
+					return fmt.Errorf("invalid activates-on value %q: slot is also activatable on app %q", slot.Name, otherApp.Name)
+				}
+			}
+		}
+	}
+
+	return nil
+}
+
 // appContentWhitelist is the whitelist of legal chars in the "apps"
 // section of snap.yaml. Do not allow any of [',",`] here or snap-exec
 // will get confused. chainContentWhitelist is the same, but for the
@@ -586,6 +707,19 @@ func ValidateApp(app *AppInfo) error {
 		return fmt.Errorf(`"daemon" field contains invalid value %q`, app.Daemon)
 	}
 
+	switch app.DaemonScope {
+	case "":
+		if app.Daemon != "" {
+			return fmt.Errorf(`"daemon-scope" must be set for daemons`)
+		}
+	case SystemDaemon, UserDaemon:
+		if app.Daemon == "" {
+			return fmt.Errorf(`"daemon-scope" can only be set for daemons`)
+		}
+	default:
+		return fmt.Errorf(`invalid "daemon-scope": %q`, app.DaemonScope)
+	}
+
 	// Validate app name
 	if !ValidAppName(app.Name) {
 		return fmt.Errorf("cannot have %q as app name - use letters, digits, and dash as separator", app.Name)
@@ -626,6 +760,10 @@ func ValidateApp(app *AppInfo) error {
 		}
 	}
 
+	if err := validateAppActivatesOn(app); err != nil {
+		return err
+	}
+
 	if err := validateAppRestart(app); err != nil {
 		return err
 	}
@@ -651,12 +789,22 @@ func ValidateApp(app *AppInfo) error {
 	default:
 		return fmt.Errorf(`"refresh-mode" field contains invalid value %q`, app.RefreshMode)
 	}
+	// validate install-mode
+	switch app.InstallMode {
+	case "", "enable", "disable":
+		// valid
+	default:
+		return fmt.Errorf(`"install-mode" field contains invalid value %q`, app.InstallMode)
+	}
 	if app.StopMode != "" && app.Daemon == "" {
 		return fmt.Errorf(`"stop-mode" cannot be used for %q, only for services`, app.Name)
 	}
 	if app.RefreshMode != "" && app.Daemon == "" {
 		return fmt.Errorf(`"refresh-mode" cannot be used for %q, only for services`, app.Name)
 	}
+	if app.InstallMode != "" && app.Daemon == "" {
+		return fmt.Errorf(`"install-mode" cannot be used for %q, only for services`, app.Name)
+	}
 
 	return validateAppTimer(app)
 }
@@ -727,6 +875,72 @@ func (layout *Layout) constraint() LayoutConstraint {
 	return mountedTree(path)
 }
 
+// layoutRejectionList contains directories that cannot be used as layout
+// targets. Nothing there, or underneath can be replaced with $SNAP or
+// $SNAP_DATA, or $SNAP_COMMON content, even from the point of view of a single
+// snap.
+var layoutRejectionList = []string{
+	// Special locations that need to retain their properties:
+
+	// The /dev directory contains essential device nodes and there's no valid
+	// reason to allow snaps to replace it.
+	"/dev",
+	// The /proc directory contains essential process meta-data and
+	// miscellaneous kernel configuration parameters and there is no valid
+	// reason to allow snaps to replace it.
+	"/proc",
+	// The /sys directory exposes many kernel internals, similar to /proc and
+	// there is no known reason to allow snaps to replace it.
+	"/sys",
+	// The media directory is mounted with bi-directional mount event sharing.
+	// Any mount operations there are reflected in the host's view of /media,
+	// which may be either itself or /run/media.
+	"/media",
+	// The /run directory contains various ephemeral information files or
+	// sockets used by various programs. Providing view of the true /run allows
+	// snap applications to be integrated with the rest of the system and
+	// therefore snaps should not be allowed to replace it.
+	"/run",
+	// The /tmp directory contains a private, per-snap, view of /tmp and
+	// there's no valid reason to allow snaps to replace it.
+	"/tmp",
+	// The /var/lib/snapd directory contains essential snapd state and is
+	// sometimes consulted from inside the mount namespace.
+	"/var/lib/snapd",
+
+	// Locations that may be used to attack the host:
+
+	// The firmware is sometimes loaded on demand by the kernel, in response to
+	// a process performing generic I/O to a specific device. In that case the
+	// mount namespace of the process is searched, by the kernel, for the
+	// firmware. Therefore firmware must not be replaceable to prevent
+	// malicious firmware from attacking the host.
+	"/lib/firmware",
+	// Similarly the kernel will load modules and the modules should not be
+	// something that snaps can tamper with.
+	"/lib/modules",
+
+	// Locations that store essential data:
+
+	// The /var/snap directory contains system-wide state of particular snaps
+	// and should not be replaced as it would break content interface
+	// connections that use $SNAP_DATA or $SNAP_COMMON.
+	"/var/snap",
+	// The /home directory contains user data, including $SNAP_USER_DATA,
+	// $SNAP_USER_COMMON and should be disallowed for the same reasons as
+	// /var/snap.
+	"/home",
+
+	// Locations that should be pristine to avoid confusion.
+
+	// There's no known reason to allow snaps to replace things there.
+	"/boot",
+	// The lost+found directory is used by fsck tools to link lost blocks back
+	// into the filesystem tree. Using layouts for this element is just
+	// confusing and there is no valid reason to allow it.
+	"/lost+found",
+}
+
 // ValidateLayout ensures that the given layout contains only valid subset of constructs.
 func ValidateLayout(layout *Layout, constraints []LayoutConstraint) error {
 	si := layout.Snap
@@ -750,7 +964,7 @@ func ValidateLayout(layout *Layout, constraints []LayoutConstraint) error {
 		return fmt.Errorf("layout %q uses invalid mount point: must be absolute and clean", layout.Path)
 	}
 
-	for _, path := range []string{"/proc", "/sys", "/dev", "/run", "/boot", "/lost+found", "/media", "/var/lib/snapd", "/var/snap", "/lib/firmware", "/lib/modules"} {
+	for _, path := range layoutRejectionList {
 		// We use the mountedTree constraint as this has the right semantics.
 		if mountedTree(path).IsOffLimits(mountPoint) {
 			return fmt.Errorf("layout %q in an off-limits area", layout.Path)
@@ -860,3 +1074,90 @@ func ValidateCommonIDs(info *Info) error {
 	}
 	return nil
 }
+
+func ValidateSystemUsernames(info *Info) error {
+	for username := range info.SystemUsernames {
+		if !osutil.IsValidUsername(username) {
+			return fmt.Errorf("invalid system username %q", username)
+		}
+	}
+	return nil
+}
+
+// NeededDefaultProviders returns a map keyed by the names of all
+// default-providers for the content plugs that the given snap.Info
+// needs. The map values are the corresponding content tags.
+func NeededDefaultProviders(info *Info) (providerSnapsToContentTag map[string][]string) {
+	providerSnapsToContentTag = make(map[string][]string)
+	for _, plug := range info.Plugs {
+		gatherDefaultContentProvider(providerSnapsToContentTag, plug)
+	}
+	return providerSnapsToContentTag
+}
+
+// ValidateBasesAndProviders checks that all bases/default-providers are part of the seed
+func ValidateBasesAndProviders(snapInfos []*Info) []error {
+	all := naming.NewSnapSet(nil)
+	for _, info := range snapInfos {
+		all.Add(info)
+	}
+
+	var errs []error
+	for _, info := range snapInfos {
+		// ensure base is available
+		if info.Base != "" && info.Base != "none" {
+			if !all.Contains(naming.Snap(info.Base)) {
+				errs = append(errs, fmt.Errorf("cannot use snap %q: base %q is missing", info.InstanceName(), info.Base))
+			}
+		}
+		// ensure core is available
+		if info.Base == "" && info.SnapType == TypeApp && info.InstanceName() != "snapd" {
+			if !all.Contains(naming.Snap("core")) {
+				errs = append(errs, fmt.Errorf(`cannot use snap %q: required snap "core" missing`, info.InstanceName()))
+			}
+		}
+		// ensure default-providers are available
+		for dp := range NeededDefaultProviders(info) {
+			if !all.Contains(naming.Snap(dp)) {
+				errs = append(errs, fmt.Errorf("cannot use snap %q: default provider %q is missing", info.InstanceName(), dp))
+			}
+		}
+	}
+	return errs
+}
+
+var isValidLinksKey = regexp.MustCompile("^[a-zA-Z](?:-?[a-zA-Z0-9])*$").MatchString
+var validLinkSchemes = []string{"http", "https"}
+
+// ValidateLinks checks that links entries have valid keys and values that can be parsed as URLs or are email addresses possibly prefixed with mailto:.
+func ValidateLinks(links map[string][]string) error {
+	for linksKey, linksValues := range links {
+		if linksKey == "" {
+			return fmt.Errorf("links key cannot be empty")
+		}
+		if !isValidLinksKey(linksKey) {
+			return fmt.Errorf("links key is invalid: %s", linksKey)
+		}
+		if len(linksValues) == 0 {
+			return fmt.Errorf("%q links cannot be specified and empty", linksKey)
+		}
+		for _, link := range linksValues {
+			if link == "" {
+				return fmt.Errorf("empty %q link", linksKey)
+			}
+			u, err := url.Parse(link)
+			if err != nil {
+				return fmt.Errorf("invalid %q link %q", linksKey, link)
+			}
+			if u.Scheme == "" || u.Scheme == "mailto" {
+				// minimal check
+				if !strings.Contains(link, "@") {
+					return fmt.Errorf("invalid %q email address %q", linksKey, link)
+				}
+			} else if !strutil.ListContains(validLinkSchemes, u.Scheme) {
+				return fmt.Errorf("%q link must have one of http|https schemes or it must be an email address: %q", linksKey, link)
+			}
+		}
+	}
+	return nil
+}
